@@ -9,14 +9,16 @@ import hashlib
 from typing import Any
 
 from xpra.util.objects import typedict
-from xpra.util.str_fn import csv, bytestostr
+from xpra.util.str_fn import csv
 from xpra.util.env import envbool
 from xpra.os_util import gi_import
-from xpra.gtk.keymap import get_gtk_keymap
+from xpra.gtk.keymap import get_gtk_keymap, get_default_keymap
 from xpra.gtk.error import xsync, xlog
-from xpra.keyboard.mask import DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names
+from xpra.keyboard.mask import (
+    DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names,
+)
 from xpra.server.keyboard_config_base import KeyboardConfigBase
-from xpra.x11.gtk_x11.keys import grok_modifier_map
+from xpra.x11.gtk.keys import grok_modifier_map
 from xpra.x11.xkbhelper import (
     do_set_keymap, set_all_keycodes, set_keycode_translation,
     get_modifiers_from_meanings, get_modifiers_from_keycodes,
@@ -49,7 +51,7 @@ ALL_X11_MODIFIERS: dict[str, int] = {
 }
 
 
-def kmlog(keyname, msg, *args) -> None:
+def kmlog(keyname: str, msg: str, *args) -> None:
     """ logs at info level is we are debugging this keyname """
     if keyname in DEBUG_KEYSYMS:
         lfn = log.info
@@ -65,25 +67,24 @@ class KeyboardConfig(KeyboardConfigBase):
         self.query_struct = None
         self.modifier_map: dict[str, int] = {}
         self.mod_meanings = {}
-        self.mod_managed = []
-        self.mod_pointermissing = []
-        self.mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
+        self.mod_managed: list[str] = []
+        self.mod_pointermissing: list[str] = []
+        self.mod_nuisance: set[str] = set(DEFAULT_MODIFIER_NUISANCE)
         self.keycodes = ()
         self.x11_keycodes = []
-        self.layout = None
-        self.variant = None
-        self.options = None
+        self.layout = ""
+        self.variant = ""
+        self.options = ""
         self.layout_groups = False
 
         # this is shared between clients!
         self.keys_pressed = {}
         # these are derived by calling set_keymap:
-        self.keynames_for_mod = {}
+        self.keynames_for_mod: dict[str, list[str]] = {}
         self.keycode_translation = {}
-        self.keycodes_for_modifier_keynames = {}
-        self.modifier_client_keycodes = {}
+        self.keycodes_for_modifier_keynames: dict[str, list[int]] = {}
+        self.modifier_client_keycodes: dict[str, list] = {}
         self.compute_modifier_map()
-        self.modifiers_filter = []
         self.keycode_mappings = {}
 
     def __repr__(self):
@@ -115,9 +116,7 @@ class KeyboardConfig(KeyboardConfigBase):
                 kminfo[i] = (keyval, name, keycode, group, level)
                 i += 1
         # modifiers:
-        modinfo: dict[str, dict[str, Any]] = {
-            "filter": self.modifiers_filter,
-        }
+        modinfo: dict[str, dict[str, Any]] = {}
         if self.modifier_client_keycodes:
             for mod, keys in self.modifier_client_keycodes.items():
                 modinfo.setdefault(mod, {})["client_keys"] = keys
@@ -150,7 +149,7 @@ class KeyboardConfig(KeyboardConfigBase):
         log("keyboard info: %s", info)
         return info
 
-    def parse_options(self, props) -> int:
+    def parse_options(self, props: typedict) -> int:
         """ used by both process_hello and process_keymap
             to set the keyboard attributes """
         super().parse_options(props)
@@ -181,7 +180,7 @@ class KeyboardConfig(KeyboardConfigBase):
         log("assign_keymap_options(..) modified %s", modded)
         return len(modded)
 
-    def parse_layout(self, props) -> None:
+    def parse_layout(self, props: typedict) -> None:
         """ used by both process_hello and process_keymap """
         keymap_dict = typedict(props.dictget("keymap") or {})
         self.layout = keymap_dict.strget("layout", "us")
@@ -234,8 +233,10 @@ class KeyboardConfig(KeyboardConfigBase):
                 if MAP_MISSING_MODIFIERS:
                     map_missing_modifiers(self.keynames_for_mod)
             else:
-                log.info("client did not supply any modifier definitions")
-                self.keynames_for_mod = {}
+                log.info("client did not supply any modifier definitions, using defaults")
+                self.keynames_for_mod = get_modifiers_from_meanings(DEFAULT_MODIFIER_MEANINGS)
+                if MAP_MISSING_MODIFIERS:
+                    map_missing_modifiers(self.keynames_for_mod)
         log("compute_modifiers() keynames_for_mod=%s", self.keynames_for_mod)
         log("compute_modifiers() keycodes_for_modifier_keynames=%s", self.keycodes_for_modifier_keynames)
         log("compute_modifiers() mod_meanings=%s", self.mod_meanings)
@@ -243,14 +244,13 @@ class KeyboardConfig(KeyboardConfigBase):
     def compute_modifier_keynames(self) -> None:
         self.keycodes_for_modifier_keynames = {}
         self.mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
-        display = Gdk.Display.get_default()
-        keymap = Gdk.Keymap.get_for_display(display)
+        keymap = get_default_keymap()
         if self.keynames_for_mod:
             for modifier, keynames in self.keynames_for_mod.items():
                 for keyname in keynames:
                     if keyname in DEFAULT_MODIFIER_NUISANCE_KEYNAMES:
                         self.mod_nuisance.add(modifier)
-                    keyval = Gdk.keyval_from_name(bytestostr(keyname))
+                    keyval = Gdk.keyval_from_name(keyname)
                     if keyval == 0:
                         log.error("Error: no keyval found for keyname '%s' (modifier '%s')", keyname, modifier)
                         continue
@@ -305,7 +305,7 @@ class KeyboardConfig(KeyboardConfigBase):
 
     def compute_modifier_map(self) -> None:
         with xlog:
-            self.modifier_map = grok_modifier_map(Gdk.Display.get_default(), self.mod_meanings)
+            self.modifier_map = grok_modifier_map(self.mod_meanings)
         log("modifier_map(%s)=%s", self.mod_meanings, self.modifier_map)
 
     def is_modifier(self, keycode: int) -> bool:
@@ -316,7 +316,7 @@ class KeyboardConfig(KeyboardConfigBase):
         log("is_modifier(%s) not found", keycode)
         return False
 
-    def set_layout(self, layout: str, variant: str, options) -> bool:
+    def set_layout(self, layout: str, variant: str, options: str) -> bool:
         log("set_layout(%s, %s, %s)", layout, variant, options)
         if layout == "en":
             log.warn("Warning: invalid keyboard layout name '%s', using 'us' instead", layout)
@@ -336,6 +336,7 @@ class KeyboardConfig(KeyboardConfigBase):
         if translate_only:
             self.keycode_translation = set_keycode_translation(self.x11_keycodes, self.keycodes)
             self.add_gtk_keynames()
+            self.add_loose_matches()
             self.compute_modifiers()
             self.compute_modifier_keynames()
             self.compute_client_modifier_keycodes()
@@ -388,6 +389,7 @@ class KeyboardConfig(KeyboardConfigBase):
             log("keyname_for_mod=%s", self.keynames_for_mod)
             clean_keyboard_state()
             self.update_keycode_mappings()
+            self.add_loose_matches()
 
     def add_gtk_keynames(self) -> None:
         # add the keynames we find via gtk
@@ -400,6 +402,12 @@ class KeyboardConfig(KeyboardConfigBase):
                 if keyname in DEBUG_KEYSYMS:
                     log.info("add_gtk_keynames: %s=%s", keyname, keycode)
 
+    def add_loose_matches(self) -> None:
+        # add lowercase versions of all keynames
+        for key, keycode in tuple(self.keycode_translation.items()):
+            if isinstance(key, str) and not key.islower():
+                self.keycode_translation[key.lower()] = keycode
+
     def set_default_keymap(self) -> None:
         """
         assign a default keymap based on the current X11 server keymap
@@ -451,8 +459,8 @@ class KeyboardConfig(KeyboardConfigBase):
     def update_keycode_mappings(self) -> None:
         self.keycode_mappings = get_keycode_mappings()
 
-    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool, modifiers, keyval,
-                       keystr: str, group: int):
+    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool, modifiers, keyval: int,
+                       keystr: str, group: int) -> tuple[int, int]:
         if not self.enabled:
             kmlog(keyname, "ignoring keycode since keyboard is turned off")
             return -1, group
@@ -467,7 +475,7 @@ class KeyboardConfig(KeyboardConfigBase):
         return self.find_matching_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)
 
     def find_matching_keycode(self, client_keycode: int, keyname: str,
-                              pressed: bool, modifiers, keyval, keystr: str, group) -> tuple[int, int]:
+                              pressed: bool, modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
         """
         from man xmodmap:
         The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
@@ -478,14 +486,14 @@ class KeyboardConfig(KeyboardConfigBase):
         the second with Shift, the third when the Mode_switch key is used with this key and
         the fourth when both the Mode_switch and Shift keys are used.
         """
-        keycode = None
+        keycode = -1
         rgroup = group
 
-        def klog(msg, *args):
+        def klog(msg: str, *args) -> None:
             kmlog(keyname, "do_get_keycode%s" + msg,
                   (client_keycode, keyname, pressed, modifiers, keyval, keystr, group), *args)
 
-        def kml(msg, *args):
+        def kml(msg: str, *args) -> None:
             kmlog(keyname, msg, *args)
 
         # non-native: try harder to find matching keysym
@@ -518,67 +526,67 @@ class KeyboardConfig(KeyboardConfigBase):
                     levels.append(level)
         kml("will try levels: %s", levels)
         for level in levels:
-            keycode = self.keycode_translation.get((keyname, level))
-            if keycode:
-                keysyms = self.keycode_mappings.get(keycode)
-                klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
-                level0 = levels[0]
-                uq_keysyms = set(keysyms)
-                if len(uq_keysyms) <= 1 or (len(keysyms) > level0 and keysyms[level0] == ""):
-                    # if the keysym we would match for this keycode is 'NoSymbol',
-                    # then we can probably ignore it ('NoSymbol' shows up as "")
-                    # same if there's only one actual keysym for this keycode
-                    kml("not toggling any modifiers state for keysyms=%s", keysyms)
-                    break
-
-                def toggle_modifier(mod):
-                    keynames = self.keynames_for_mod.get(mod, set())
-                    if keyname in keynames:
-                        kml("not toggling '%s' since '%s' should deal with it", mod, keyname)
-                        # the keycode we're returning is for this modifier,
-                        # assume that this will end up doing what is needed
-                        return
-                    if mod in modifiers:
-                        kml("removing '%s' from modifiers", mod)
-                        modifiers.remove(mod)
-                    else:
-                        kml("adding '%s' to modifiers", mod)
-                        modifiers.append(mod)
-
-                # keypad overrules shift state (see #2702):
-                if keyname.startswith("KP_"):
-                    if numlock_modifier and not numlock:
-                        toggle_modifier(numlock_modifier)
-                elif (level & 1) ^ shift:
-                    # shift state does not match
-                    toggle_modifier("shift")
-                if int(bool(level & 2)) ^ mode:
-                    # try to set / unset mode:
-                    for mod, keynames in self.keynames_for_mod.items():
-                        if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
-                            # found mode switch modified
-                            toggle_modifier(mod)
-                            break
-                rgroup = level // 4
-                if rgroup != group:
-                    kml("switching group from %i to %i", group, rgroup)
+            keycode = self.keycode_translation.get((keyname, level), -1)
+            if keycode < 0:
+                continue
+            keysyms = self.keycode_mappings.get(keycode)
+            klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
+            level0 = levels[0]
+            uq_keysyms = set(keysyms)
+            if len(uq_keysyms) <= 1 or (len(keysyms) > level0 and keysyms[level0] == ""):
+                # if the keysym we would match for this keycode is 'NoSymbol',
+                # then we can probably ignore it ('NoSymbol' shows up as "")
+                # same if there's only one actual keysym for this keycode
+                kml("not toggling any modifiers state for keysyms=%s", keysyms)
                 break
+
+            def toggle_modifier(mod: str) -> None:
+                keynames = self.keynames_for_mod.get(mod, set())
+                if keyname in keynames:
+                    kml("not toggling '%s' since '%s' should deal with it", mod, keyname)
+                    # the keycode we're returning is for this modifier,
+                    # assume that this will end up doing what is needed
+                    return
+                if mod in modifiers:
+                    kml(f"removing {mod!r} from modifiers")
+                    modifiers.remove(mod)
+                else:
+                    kml(f"adding {mod} to modifiers")
+                    modifiers.append(mod)
+
+            # keypad overrules shift state (see #2702):
+            if keyname.startswith("KP_"):
+                if numlock_modifier and not numlock:
+                    toggle_modifier(numlock_modifier)
+            elif (level & 1) ^ shift:
+                # shift state does not match
+                toggle_modifier("shift")
+            if int(bool(level & 2)) ^ mode:
+                # try to set / unset mode:
+                for mod, keynames in self.keynames_for_mod.items():
+                    if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
+                        # found mode switch modified
+                        toggle_modifier(mod)
+                        break
+            rgroup = level // 4
+            if rgroup != group:
+                kml("switching group from %i to %i", group, rgroup)
+            break
         # this should not find anything new?:
-        if keycode is None:
+        if keycode < 0:
             keycode = self.keycode_translation.get(keyname, -1)
             klog("=%i, %i (keyname translation)", keycode, rgroup)
-        if keycode is None:
+        if keycode < 0:
             # last resort, find using the keyval:
-            display = Gdk.Display.get_default()
-            keymap = Gdk.Keymap.get_for_display(display)
+            keymap = get_default_keymap()
             b, entries = keymap.get_entries_for_keyval(keyval)
             if b and entries:
                 for entry in entries:
                     if entry.group == group:
                         return entry.keycode, entry.group
-        return keycode or 0, rgroup or 0
+        return keycode, rgroup
 
-    def get_current_mask(self) -> list:
+    def get_current_mask(self) -> list[str]:
         # this is only called from an existing xsync context in this module
         return mask_to_names(X11Keyboard.query_mask(), self.modifier_map)
 
@@ -608,19 +616,18 @@ class KeyboardConfig(KeyboardConfigBase):
             return
         if ignored_modifier_keynames is None:
             # this is not a keyboard event, ignore modifiers in "mod_pointermissing"
-            def is_ignored(modifier, _modifier_keynames):
+            def is_ignored(modifier, _modifier_keynames) -> bool:
                 return modifier in (self.mod_pointermissing or [])
         else:
             # keyboard event: ignore the keynames specified
             # (usually the modifier key being pressed/unpressed)
-            def is_ignored(_modifier, modifier_keynames):
+            def is_ignored(_modifier, modifier_keynames) -> bool:
                 return len(set(modifier_keynames or []) & set(ignored_modifier_keynames or [])) > 0
 
-        def filtered_modifiers_set(modifiers):
-            m = set()
+        def filtered_modifiers_set(modifiers) -> set[str]:
+            m: set[str] = set()
             mm = self.mod_managed or ()
             for modifier in modifiers:
-                modifier = bytestostr(modifier)
                 if modifier in mm:
                     log("modifier is server managed: %s", modifier)
                     continue
@@ -632,10 +639,9 @@ class KeyboardConfig(KeyboardConfigBase):
             log("filtered_modifiers_set(%s)=%s", modifiers, m)
             return m
 
-        def change_mask(modifiers, press, info):
+        def change_mask(modifiers: set[str], press: bool, info: str) -> list[str]:
             failed = []
             for modifier in modifiers:
-                modifier = bytestostr(modifier)
                 if modifier not in self.keynames_for_mod:
                     log.error(f"Error: unknown modifier {modifier!r}")
                     log.error(" known modifiers: %s", csv(self.keynames_for_mod.keys()))
